
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>

#include "parser.h"
#include "table.h"
#include "main.h"
#include "gccdeps.h"
#include "command.h"

#include "schedule.h"
 typedef struct cmdlist {
			struct cmdlist *prev,*next;
			strtab_t fields;
		} cmdlist_t;
 static  void print_outputs(strtab_t *c) ; static  void
create_command_and_output_list() ; static  int command_is_ready(strtab_t *c) ; static  int command_is_clean(strtab_t *c) ; static  void cmd_destroy(strtab_t *C) ; static  void
execute_command(strtab_t *c) ; static  char*
prepare_string(strtab_t *c,char *field) ; static  int 
prep_string(arg_t *arg, strtab_t *c, char *result) ; static  int put_string(char *result, int len, char *src) ; static  int put_token(char *result, int len, int tk) ; static  unsigned
modtime_input(char *fn) ; static  unsigned
modtime_output(char *fn) ; static  unsigned
modtime_weak_input(char *fn) ; static  unsigned modtime(char *fn) ; static  void modtime_force(char *fn,unsigned V) ; static  void modtime_update(char *fn) ;
 static cmdlist_t cmds = {&cmds, &cmds};  static  strtab_t outputs; static  strtab_t modtimes;




#define MAXUNSIGNED 0xffffffffu

 void
run_commands() {
	int change=1;
	create_command_and_output_list();
	while(cmds.prev!=&cmds && change) {
		cmdlist_t *c;
		change= 0;
		  {
	void *N3;
	for(c=(&cmds)->next;c!=(&cmds);c=N3)  {
		N3= c->next;

				if (!command_is_ready(&c->fields))
					continue;
				execute_command(&c->fields);
				 {
	(c)->prev->next= (c)->next; 
	(c)->next->prev= (c)->prev; 
}


				cmd_destroy(&c->fields);
				free(c);
				change=1;
		
    } }


	}
	if (cmds.prev!=&cmds) {
		cmdlist_t *c;
		fprintf(stderr,"jj: there are one or more cyclic dependencies\n");
		fprintf(stderr,"jj: regarding the generation of following files:\n");
		  {
	void *N6;
	for(c=(&cmds)->next;c!=(&cmds);c=N6)  {
		N6= c->next;

			print_outputs(&c->fields);
		
    } }


		fprintf(stderr,"jj: I don't know what these cycles are\n");
		fprintf(stderr,"jj: but I'm almost pretty sure that\n");
		fprintf(stderr,"jj: they are there. \n");
		exit(1);
	}	
}

 static  void print_outputs(strtab_t *c) {
	arg_t *out;
	token_t *T;
	out= strtab_find(c,"outputs");
	if (out) {
		fprintf(stderr,"jj: ");
		  {
	void *N9;
	for(T=(&out->tokens)->next;T!=(&out->tokens);T=N9)  {
		N9= T->next;

			fprintf(stderr, "%s ",T->data);
		
    } }


		fprintf(stderr,"\n");
	}
}


 static  void
create_command_and_output_list() {
	while(1) {
		strtab_t cmd;
		cmdlist_t *c;
		arg_t *O;
		if (command_get_one(&cmd))
			return ;
		
	(c)= (typeof(c)) malloc(sizeof(*(c)));


		
		
		c->fields= cmd;
		 {
	(c)->prev= ((&cmds))->prev; 
	(c)->next= ((&cmds)); 
	((&cmds))->prev->next= (c); 
	((&cmds))->prev= (c); 
}


		
		O= strtab_find(&cmd,"outputs");
		if (!O)
			continue;
		token_t *T;
		  {
	void *N16;
	for(T=(&O->tokens)->next;T!=(&O->tokens);T=N16)  {
		N16= T->next;

			if (T->kind!='W') continue;
			if (!strtab_find(&outputs,T->data))
				strtab_insert(&outputs, T->data, (void*) 1);
		
    } }


	}
}

 static  int command_is_ready(strtab_t *c) {
	token_t *T;
	arg_t *triggers;
	triggers= strtab_find(c, "triggers");
	if (triggers) {
		  {
	void *N19;
	for(T=(&triggers->tokens)->next;T!=(&triggers->tokens);T=N19)  {
		N19= T->next;

			if (T->kind!='W')
				continue;
			if (strtab_find(&outputs,T->data))
				return 0;
		
    } }


	}
	 {
	arg_t *out;
	out= strtab_find(c, "outputs");
	if (out) {
		int not_ready= 0;
		  {
	void *N67;
	for(T=(&out->tokens)->next;T!=(&out->tokens);T=N67)  {
		N67= T->next;

			char *gdeps;
			char *trigger;
			if (T->kind!='W') continue;
			gdeps= get_gcc_deps(T->data);
			if (!gdeps)
				continue;
			
{
	
	
	char *_src = gdeps;
	char _div= ' ';
	char *_s, *_e;

	_s= _src;
	_e= _src;

_str_loop_68:
	while(*_e && *_e!=_div) 
		_e++;

	if (_e==_s) {
		trigger= 0;
	} else {
		int _L= _e-_s;
		trigger= (char*) malloc(_L+1);
		memcpy(trigger, _s, _L);
		trigger[_L]= 0;
	}
	{ /* brace for code, not necessary, but maybe
	the user has some declarations eh? */


				if (trigger) {
					if (strtab_find(&outputs,trigger)) {
						not_ready= 1;
						
	{ if (trigger) free(trigger); goto _str_exit_68; }


					}
				}
			
	}
_str_cont_68:
	if (trigger) free(trigger);
	if (!*_e) goto _str_exit_68;
	_e++;
	if (*_e) {
		_s= _e;
		goto _str_loop_68;
	}
_str_exit_68:
	_e= _s; /* to keep the compiler happy */
}


			if (not_ready)
				break;
		
    } }


		if (not_ready)
			return 0;
	}
}



	return 1;
}

 static  int command_is_clean(strtab_t *c) {
	unsigned min_o= MAXUNSIGNED,max_i=0;
	arg_t *A;
	token_t *T;

	A= strtab_find(c,"triggers");
	if (A) {
		  {
	void *N23;
	for(T=(&A->tokens)->next;T!=(&A->tokens);T=N23)  {
		N23= T->next;

			if (T->kind!='W') continue;
			unsigned mt= modtime_input(T->data);
			if (mt>max_i)
				max_i= mt;
		
    } }


	}

	 {
	A= strtab_find(c,"outputs");
	if (A) {
		  {
	void *N73;
	for(T=(&A->tokens)->next;T!=(&A->tokens);T=N73)  {
		N73= T->next;

			char *gdeps;
			char *trigger;
			if (T->kind!='W') continue;
			gdeps= get_gcc_deps(T->data);
			if (!gdeps)
				continue;
			
{
	
	
	char *_src = gdeps;
	char _div= ' ';
	char *_s, *_e;

	_s= _src;
	_e= _src;

_str_loop_74:
	while(*_e && *_e!=_div) 
		_e++;

	if (_e==_s) {
		trigger= 0;
	} else {
		int _L= _e-_s;
		trigger= (char*) malloc(_L+1);
		memcpy(trigger, _s, _L);
		trigger[_L]= 0;
	}
	{ /* brace for code, not necessary, but maybe
	the user has some declarations eh? */


				if (trigger) {
					unsigned mt= modtime_weak_input(trigger);
					if (mt>max_i)
						max_i= mt;
				}
			
	}
_str_cont_74:
	if (trigger) free(trigger);
	if (!*_e) goto _str_exit_74;
	_e++;
	if (*_e) {
		_s= _e;
		goto _str_loop_74;
	}
_str_exit_74:
	_e= _s; /* to keep the compiler happy */
}


		
    } }


	}
}

	A= strtab_find(c, "outputs");
	if (A) {
		  {
	void *N26;
	for(T=(&A->tokens)->next;T!=(&A->tokens);T=N26)  {
		N26= T->next;

			if (T->kind!='W') continue;
			unsigned mt= modtime_output(T->data);
			if (mt<min_o)
				min_o= mt;
		
    } }


	}
	return max_i < min_o;
}

 static  void cmd_destroy(strtab_t *C) {
	arg_t *A;
	return ;
	while((A=strtab_destroy(C)))
		arg_destroy_deep(A);
	free(C);
}

 static  void
execute_command(strtab_t *c) {
	int output_modtimes=0;
	arg_t *out;
	token_t *T;
	if (!command_is_clean(c)) {
		char *action;
		char *message;

		action= prepare_string(c,"action");
		message= prepare_string(c,"message");

		
	if (!option_noprefix)
		printf("jj: ");
	if (!action && !message) {
		printf("'Executing' a mystery command with no\n");
		if (!option_noprefix)
			printf("jj: ");
		printf("action or message in it..\n");
	} else {
		if (option_printmsg) {
			if (message)
				printf("%s\n",message);
			else
				printf("%s [NO MESSAGE DEFINED]\n", action);
		} else {
			if (action)
				printf("%s\n",action);
			else
				printf("%s [NO ACTION DEFINED]\n", message);
		}
	}

		if (option_norun) {
			output_modtimes= 'M';
		} else {
			if (system(action))
				exit(fprintf(stderr,"jj: child failed. exiting.\n"));
			output_modtimes= 'C';
		}
	}
	out= strtab_find(c,"outputs");
	if (out)
		  {
	void *N31;
	for(T=(&out->tokens)->next;T!=(&out->tokens);T=N31)  {
		N31= T->next;

				if (T->kind!='W')
					continue;
				if (output_modtimes=='M')
					modtime_force(T->data, MAXUNSIGNED);
				else if (output_modtimes=='C')
					modtime_update(T->data);
				strtab_remove(&outputs, T->data);
		
    } }


}



 static  char*
prepare_string(strtab_t *c,char *field) {
	arg_t *A;
	char *result;
	int len;
	A= strtab_find(c, field);
	if (!A)
		return 0;
	len= prep_string(A, c, 0);
	if (len<2)
		return 0;
	result= malloc(len);
	prep_string(A,c,result);
	return result;
}


 static  int 
prep_string(arg_t *arg, strtab_t *c, char *result) {
	token_t *T;
	int len=0;
	  {
	void *N35;
	for(T=(&arg->tokens)->next;T!=(&arg->tokens);T=N35)  {
		N35= T->next;

		if (T->kind!='W')  {
			len= put_token(result,len,T->kind);
		} else if (T->data[0]!='$') {
			len= put_string( result, len, T->data );
		} else {
			arg_t *A;
			A= strtab_find(c, T->data+1);
			if (A) {
				token_t *S;
				  {
	void *N36;
	for(S=(&A->tokens)->next;S!=(&A->tokens);S=N36)  {
		N36= S->next;

					if (S->kind=='W')
						len= put_string(result,len,S->data);
					else
						len= put_token(result,len,S->kind);
				
    } }


			} else {
				len= put_string( result, len, T->data );
			}
		}
	
    } }


	if (result)
		result[len]= 0;
	return len+1;
}

 static  int put_string(char *result, int len, char *src) {
	int L= strlen(src);
	if (result) {
		memcpy(result+len, src, L);
		result[len+L]= ' ';
	}
	len+= L+1;
	return len;
}

 static  int put_token(char *result, int len, int tk) {
	char buf[2];
	buf[0]= tk;
	buf[1]= 0;
	return put_string(result, len, buf);
}

 static  unsigned
modtime_input(char *fn) {
	unsigned mt;
	mt= modtime(fn);
	if (!mt)
		exit(fprintf(stderr,"can not stat %s: %s\n",fn,strerror(errno)));
	return mt;
}

/* This function is used for checking output files. 
   If the file doesn`t exist, it`s assumed to have a 
   very old timestamp, i.e. 0 == 1970 */
 static  unsigned
modtime_output(char *fn) {
	return modtime(fn);
}

/* weak inputs are just like outputs. if they don`t exist,
   they simply cause a generation. hopefully, this generation
   will remove them from the weak input pool (i.e. outdated -MMD). */
 static  unsigned
modtime_weak_input(char *fn) {
	return modtime_output(fn);
}

 static  unsigned modtime(char *fn) {
	unsigned V;
	/* a cache is mandatory. sometimes we force
	modtimes on files and sometimes the file does not
	fucking exist in the first place */
	V= (unsigned) strtab_find(&modtimes,fn);
	if (V)
		return V;
	struct stat sbuf;
	if (stat(fn,&sbuf))
		return 0;
	V= sbuf.st_mtime;
	if (!V)
		V++;
	strtab_insert(&modtimes, fn, (void*) V);
	return V;
}

 static  void modtime_force(char *fn,unsigned V) {
	if (strtab_find(&modtimes,fn))
		strtab_remove(&modtimes,fn);
	strtab_insert(&modtimes, fn, (void*) V);
}



 static  void modtime_update(char *fn) {
	unsigned V;
	V= (unsigned) strtab_find(&modtimes,fn);
	if (V)
		strtab_remove(&modtimes,fn);
	struct stat sbuf;
	if (stat(fn,&sbuf)) {
		strtab_insert(&modtimes,fn, (void*) MAXUNSIGNED);
		return ;
	}
	V= sbuf.st_mtime;
	if (!V)
		V++;
	strtab_insert(&modtimes, fn, (void*) V);
}

 

  
 

 





