
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "parser.h"
#include "table.h"
#include "functions.h"
#include "string.h"
#include "print.h"

#include "command.h"
 typedef struct entry entry_t;  struct entry {
	int id;
	strtab_t fields;
	
  entry_t *left, *right;
  int height;

}; typedef struct {
	entry_t *root;
} cmdtab_t;
 static  void
cmdtab_init(cmdtab_t *t) ; static  void*
cmdtab_find(cmdtab_t *t,entry_t *key) ; static  int
cmdtab_insert(cmdtab_t *t,entry_t *node) ; static  entry_t*
cmdtab_insert_node(entry_t *t,entry_t *n) ; static  entry_t*
cmdtab_balance(entry_t *t) ; static  entry_t*
cmdtab_remove(cmdtab_t *t, entry_t *key) ; static  entry_t*
cmdtab_remove_node(entry_t *t,entry_t *node) ; static  entry_t*
cmdtab_rotate_right(entry_t *g) ; static  entry_t*
cmdtab_rotate_left(entry_t *g) ; static  entry_t*
cmdtab_remove_max(entry_t *t,entry_t **pl) ; static  entry_t*
cmdtab_remove_min(entry_t *t,entry_t **pl) ; static  void
cmdtab_calc_height(entry_t *t) ; static  entry_t*
cmdtab_destroy(cmdtab_t *t) ; static  entry_t*
cmdtab_destroy_one_node(entry_t *node,entry_t** plink) ; static  int intcompare(entry_t *A,entry_t *B) ; static  void setfield(char *name, entry_t *E, arg_t *V) ; static  void appendfield(char *name, entry_t *E,arg_t *V) ; static  void 
expand_token_into(token_t *T, frame_t *frame, arg_t *dst) ; static  void print_entry(entry_t *E) ; static  void 
modify_command(char *target,arg_t *body, frame_t *frame) ; static  entry_t *find_by_target(entry_t *node, char *name) ;
 static  cmdtab_t commands; static  int lastid;






#define LCURLY '{'
#define RCURLY '}'










 static  void
cmdtab_init(cmdtab_t *t) { t->root= 0; }


 static  void*
cmdtab_find(cmdtab_t *t,entry_t *key) {
	entry_t *n= t->root;
	while(n) {
		int u;
		u= intcompare(key,n);
		if (u==0) return n;
		else if (u<0) n=n->left;
		else n=n->right;
	}
	return 0;
}

 static  int
cmdtab_insert(cmdtab_t *t,entry_t *node) {
	if (cmdtab_find(t,node)) {
		return 1;
	} else {
		node->height= 1;
		node->left= node->right= 0;
		t->root= cmdtab_insert_node(t->root,node);
		return 0;
	}
}

 static  entry_t*
cmdtab_insert_node(entry_t *t,entry_t *n) {
	int u;
	if (!t)
		return n;
	u= intcompare(n,t);
	if (u<0)
		t->left= cmdtab_insert_node(t->left,n);
	else
		t->right= cmdtab_insert_node(t->right,n);
	return cmdtab_balance(t);
}

#ifndef TREE_height
#define TREE_height(x) ((x)? (x)->height : 0)
#endif

 static  entry_t*
cmdtab_balance(entry_t *t) {
	int u;
	cmdtab_calc_height(t);
	u= TREE_height(t->left)-TREE_height(t->right);
	if (u>1)
		return cmdtab_rotate_right(t);
	else if (u<-1)
		return cmdtab_rotate_left(t);
	else return t;
}

 static  entry_t*
cmdtab_remove(cmdtab_t *t, entry_t *key) {
	entry_t *node;
	node= cmdtab_find(t,key);
	if (node)
		t->root= cmdtab_remove_node(t->root,node);
	return node;
}

 static  entry_t*
cmdtab_remove_node(entry_t *t,entry_t *node) {
	int u;
	u= intcompare(node,t);
	if (u<0) 
		t->left= cmdtab_remove_node(t->left,node);
	else if (u>0)
		t->right= cmdtab_remove_node(t->right,node);
	else {
		if (!t->right || !t->left) {
			if (t->right)
				return t->right;
			else
				return t->left;
		} else {
			entry_t *p;
			if (t->left->height>t->right->height)
				p= cmdtab_remove_max(t->left, &t->left);
			else 
				p= cmdtab_remove_min(t->right, &t->right);
			p->left= t->left;
			p->right= t->right;
			p->height=1;
			t= p;
		}
	}
	return cmdtab_balance(t);
}

 static  entry_t*
cmdtab_rotate_right(entry_t *g) {
	entry_t *f,*e;
	f= g->left;
	e= f->right;
	f->right= g;
	g->left= e;
	cmdtab_calc_height(g);
	cmdtab_calc_height(f);
	return f;
}

 static  entry_t*
cmdtab_rotate_left(entry_t *g) {
	entry_t *f,*e;
	f= g->right;
	e= f->left;
	f->left= g;
	g->right= e;
	cmdtab_calc_height(g);
	cmdtab_calc_height(f);
	return f;
}


 static  entry_t*
cmdtab_remove_max(entry_t *t,entry_t **pl) {
	if (t->right) {
		entry_t *z;
		z= cmdtab_remove_max(t->right,&t->right);
		cmdtab_calc_height(t);
		return z;
	}
	*pl= t->left;
	return t;
}

 static  entry_t*
cmdtab_remove_min(entry_t *t,entry_t **pl) {
	if (t->left) {
		entry_t *z;
		z= cmdtab_remove_min(t->left,&t->left);
		cmdtab_calc_height(t);
		return z;
	}
	*pl= t->right;
	return t;
}

 static  void
cmdtab_calc_height(entry_t *t) {
	int a,b;
	a= TREE_height(t->left)+1;
	b= TREE_height(t->right)+1;
	if (a>b) t->height=a;
	else t->height= b;
}

 static  entry_t*
cmdtab_destroy(cmdtab_t *t) {
	if (!t->root)
		return 0;
	return cmdtab_destroy_one_node(t->root,&t->root);
}

 static  entry_t*
cmdtab_destroy_one_node(entry_t *node,entry_t** plink) {
	if (node->left)
		return cmdtab_destroy_one_node(node->left, &node->left);
	else if (node->right)
		return cmdtab_destroy_one_node(node->right, &node->right);
	*plink= 0;
	return node;
}



 static  int intcompare(entry_t *A,entry_t *B) {
	return A->id - B->id;
}



arg_t *blt_cmdnew(arg_t *call,frame_t *frame) {
	char *ident= get_varname(call,0,"cmd.new");
	
	entry_t *E;
	
	(E)= (typeof(E)) malloc(sizeof(*(E)));


	
	memset( (E), 0, sizeof(*(E)));
	

	E->id= lastid++;
	cmdtab_insert(&commands, E);
	char buf[20];
	sprintf(buf,"%d",E->id);
	var_set(ident, arg_new_text(strdup(buf)), frame);
	return 0;
}

arg_t *blt_cmdset(arg_t *call,frame_t *frame) {
	arg_t *ident= arglist(call,0);
	arg_t *body= arglist(call,1);
	if (!ident || !body)
		wrong_args("cmd.set");
	char *itext= arg_to_string(ident,frame);
	int id= atoi(itext);
	free(itext);

	entry_t key;
	key.id= id;
	entry_t *E= cmdtab_find(&commands,&key);
	if (!E)
		runtime_error("cmd.set", "can not find command");

	token_t *T;
	int state= 'I';
	int depth=0;
	char *fieldname;
	arg_t *value;
	  {
	void *N12;
	for(T=(&body->tokens)->next;T!=(&body->tokens);T=N12)  {
		N12= T->next;

		switch(state) {
		case 'I':
			if (T->kind!='W' || T->data[0]=='$')
				runtime_error("cmd.set", "invalid field specifier");
			fieldname= T->data;
			state= 'B';
			value= arg_new();
			break;
		case 'B':
			if (T->kind=='{') /* } */ {
				state= 'L';
				depth= 1;
			} else {
				expand_token_into(T,frame,value);
				state= 'I';
				setfield(fieldname, E, value);
			}
			break;
		case 'L':
			if (T->kind=='{') /*}*/ {
				depth++;
				expand_token_into(T,frame,value);
				break;
			} 
			/*{*/ if (T->kind=='}') {
				depth--;
				if (depth) {
					expand_token_into(T,frame,value);
				} else {
					state= 'I';
					setfield(fieldname, E, value);
				}
				break;
			}
			expand_token_into(T,frame,value);
			break;
		}
	
    } }


	return 0;
}

 static  void setfield(char *name, entry_t *E, arg_t *V) {
	arg_t *current;
	current= strtab_find(&E->fields, name);
	if (current) {
		arg_destroy_deep(current);
		strtab_remove(&E->fields, name);
	}
	strtab_insert(&E->fields, name, V);
}

 static  void appendfield(char *name, entry_t *E,arg_t *V) {
	arg_t *current;
	current= strtab_find(&E->fields,name);
	if (current) {
		token_t *T;
		  {
	void *N16;
	for(T=(&V->tokens)->next;T!=(&V->tokens);T=N16)  {
		N16= T->next;

			 {
	(T)->prev->next= (T)->next; 
	(T)->next->prev= (T)->prev; 
}


			 {
	(T)->prev= ((&current->tokens))->prev; 
	(T)->next= ((&current->tokens)); 
	((&current->tokens))->prev->next= (T); 
	((&current->tokens))->prev= (T); 
}


		
    } }


	} else {
		setfield(name, E,V);
	}
}

 static  void 
expand_token_into(token_t *T, frame_t *frame, arg_t *dst) {
	arg_t *V;
	if (T->kind!='W' || T->data[0]!='$') 
		goto copy;
	V= var_get(T->data, frame);
	if (!V) 
		goto copy;
	  {
	void *N21;
	for(T=(&V->tokens)->next;T!=(&V->tokens);T=N21)  {
		N21= T->next;

		token_t *C= token_copy_deep(T);
		 {
	(C)->prev= ((&dst->tokens))->prev; 
	(C)->next= ((&dst->tokens)); 
	((&dst->tokens))->prev->next= (C); 
	((&dst->tokens))->prev= (C); 
}


	
    } }


	return ;
copy:
	T= token_copy_deep(T);
	 {
	(T)->prev= ((&dst->tokens))->prev; 
	(T)->next= ((&dst->tokens)); 
	((&dst->tokens))->prev->next= (T); 
	((&dst->tokens))->prev= (T); 
}


}

  
  
 

arg_t *dump_commands(arg_t *call,frame_t *F) {
	/* CAUTION: this is destructive man. */
	entry_t *E;
	while((E=cmdtab_destroy(&commands))) {
		print_entry(E);
	}
	return 0;
}

 static  void print_entry(entry_t *E) {
	char *key;
	printf("Command %d  ------\n", E->id);
	while((key=strtab_getleast(&E->fields))) {
		printf("[%s] ",key);
		arg_t *A=strtab_remove(&E->fields, key);
		print_arg(A);
	}
}

 int command_get_one(strtab_t *R) {
	entry_t *E;
	E= cmdtab_destroy(&commands);
	if (!E)
		return 1;
	*R= E->fields;
	free(E);
	return 0;
}


void command_unused_functions_do_not_call() {
	cmdtab_init(0);
	cmdtab_remove(0,0);
}



arg_t *blt_cmdmodify(arg_t *call,frame_t *frame) {
	arg_t *ident= arglist(call,0);
	arg_t *body= arglist(call,1);
	token_t *T;

	if (!ident || !body)
		wrong_args("cmd.modify");

	  {
	void *N35;
	for(T=(&ident->tokens)->next;T!=(&ident->tokens);T=N35)  {
		N35= T->next;

		arg_t *V;
		token_t *T2;
		if (T->kind!='W')
			continue;
		if (T->data[0]!='$') {
			modify_command(T->data, body, frame);
			continue; 
		}
		V= var_get(T->data, frame);
		if (!V)
			exit(fprintf(stderr,
				"jj: (cmd.modify) can not find variable %s\n", T->data));
		  {
	void *N36;
	for(T2=(&V->tokens)->next;T2!=(&V->tokens);T2=N36)  {
		N36= T2->next;

			if (T2->kind=='W')
				modify_command(T2->data,body,frame);
		
    } }


	
    } }


	return 0;
}

 static  void 
modify_command(char *target,arg_t *body, frame_t *frame) {
	entry_t *E= find_by_target(commands.root,target);
	if (!E) {
		fprintf(stderr,"can not find command that creates %s\n",
			target);
		exit(1);
	}

	token_t *T;
	int state= 'I';
	int depth=0;
	char *fieldname;
	arg_t *value;
	int modtype;

	  {
	void *N40;
	for(T=(&body->tokens)->next;T!=(&body->tokens);T=N40)  {
		N40= T->next;

		switch(state) {
		case 'I':
			if (T->kind!='W') 
				runtime_error("cmd.modify", "invalid modification");
			modtype= T->data[0];
			if (modtype!='+' && modtype!='=')
				runtime_error("cmd.modify", "invalid modification operator");
			state= 'F';
			break;
		case 'F':
			if (T->kind!='W') 
				runtime_error("cmd.modify", "invalid field specification");
			fieldname= T->data;
			state= 'B';
			value= arg_new();
			break;
		case 'B':
			if (T->kind==LCURLY)  {
				state= 'L';
				depth= 1;
			} else {
				expand_token_into(T,frame,value);
				state= 'I';
				if (modtype=='=')
					setfield(fieldname, E, value);
				else
					appendfield(fieldname, E, value);
			}
			break;
		case 'L':
			if (T->kind==LCURLY)  {
				depth++;
				expand_token_into(T,frame,value);
				break;
			} 
			 if (T->kind==RCURLY) {
				depth--;
				if (depth) {
					expand_token_into(T,frame,value);
				} else {
					state= 'I';
					if (modtype=='=')
						setfield(fieldname, E, value);
					else
						appendfield(fieldname,E,value);
				}
				break;
			}
			expand_token_into(T,frame,value);
			break;
		}
	
    } }


}

 static  entry_t *find_by_target(entry_t *node, char *name) {
	arg_t *out;
	entry_t *e=0;
	if (!node)
		return 0;
	out= strtab_find(&node->fields, "outputs");
	if (out) {
		token_t *T;
		  {
	void *N43;
	for(T=(&out->tokens)->next;T!=(&out->tokens);T=N43)  {
		N43= T->next;

			if (!strcmp(T->data, name))
				return node;
		
    } }


	}
	if (node->left) e= find_by_target(node->left, name);
	if (!e && node->right) e=find_by_target(node->right, name);
	return e;
}

