-[ library "functions.c", header "functions.h" ]

-[ builtin function blt_print named "print" ]
-[ builtin function string_cat named "str.cat" ]
-[ builtin function blt_dump named "dump" ]
-[ builtin function blt_eval named "eval" ]
-[ builtin function blt_set named "set" ]
-[ builtin function blt_def named "def" ]
-[ builtin function blt_strsfxrep named "rep.sfx" ]
-[ builtin function blt_comment named "comment" ]
-[ builtin function blt_listtrans named "list.trans" ]
-[ builtin function blt_include named "include" ]
-[ builtin function blt_cmdnew named "cmd.new" ]
-[ builtin function blt_cmdset named "cmd.set" ]
-[ builtin function dump_commands named "dump.cmd" ]
-[ builtin function blt_prgout named "prgout" ]
-[ builtin function blt_getenv named "env.get" ]
-[ builtin function blt_setenvkill named "env.set.kill" ]
-[ builtin function blt_depdir named "dep.dir" ]
-[ builtin function blt_cmdmodify named "cmd.modify" ]

-[ builtin function strprepend named "str.prepend" ]

-[ builtin function blt_command named "command" ]

-[private] strtab_t functions;

-[public] void
init_builtin_functions() {
	*[ builtin functions ]
}

-[public] typedef struct {
	int kind;
	char *name;
	void *func;
} function_t;


=[ builtin function $FN named $NAME ]
+[ builtin functions ]
{
        function_t *F;
        -[ptr/new F]
        F->name= strdup($NAME);
        F->func= $FN;
        F->kind= 'B';
        strtab_insert(&functions, F->name,F);
}

+[ end ]

+[private functions]
arg_t *$FN(arg_t *call,frame_t *F);
+[ end ]

=[]

-[ lheader "parser.h" ]
-[ lheader "table.h" ]
-[ header "string.h" ]
-[ header "stdlib.h" ]
-[ lheader "main.h" ]

-[public] arg_t *
exec_stmt(stmt_t *S, frame_t *frame) {
	function_t *F;

	F= strtab_find(&functions,S->func);
	if (!F)
		exit(fprintf(stderr,"can not find function %s\n",S->func));

	if (F->kind=='B')  {
		arg_t* (*func)(arg_t*,frame_t*);
		func= F->func;
		return (*func)(&S->args, frame);
	} else {
		call_user_function(S->func,&S->args,F->func,frame);
	}
	return 0;
}

-[ public] typedef struct frame {
	char* fname;
	strtab_t locals;
} frame_t;

-[public] arg_t *frame_get(frame_t *F,char *name) {
	return (arg_t*) strtab_find(&F->locals,name);
}

-[public] void frame_kill(frame_t *F,char *name) {
	arg_t *current;
	current= frame_get(F, name);
	if (current) {
		arg_destroy_deep(current);
		strtab_remove(&F->locals, name);
	}
}


-[ header "stdio.h" ]


-[public] void frame_set(frame_t *F,char *name,arg_t *value) {
	arg_t *current;
	current= frame_get(F, name);
	if (current) {
		arg_destroy_deep(current);
		strtab_remove(&F->locals, name);
	}
	
	strtab_insert(&F->locals,name, arg_copy_deep(value));
}


arg_t* blt_eval(arg_t *call, frame_t *frame) {
	arg_t *var;
	arg_t *expr;

	expr= arglist(call, 0);
	var= arglist(call,1);
	if (!expr || !var)
		wrong_args("eval");

	stmt_t *statements;
	char *err= parse_arg( expr, &statements );
	if (err)
		exit(fprintf(stderr,"%s\n",err));

	arg_t *result=0;
	stmt_t *S;
	-[ for each S in list statements, begin ]
		arg_t *ret;
		ret= exec_stmt(S,frame);
		if (ret) {
			if (result)
				arg_destroy_deep(result);
			result= ret;
		}
	-[ end ]

	if (result) {
		if (var->tokens.next->kind!='W')
			exit(fprintf(stderr,"wrong kind of argument to function `set'\n"));
		frame_set(frame, var->tokens.next->data, result);
	}

	return 0;
}

arg_t *blt_set(arg_t *call,frame_t *frame) {
	char *varname;
	arg_t *expr;

	expr= arglist(call,0);
	if (!expr)
		wrong_args("eval");
	varname= get_varname(call,1,"set");

	arg_t *value;
	value= arg_new();

	token_t *T;
	-[ for each T in list (&expr->tokens) , begin ]
		if (T->kind!='W' || T->data[0]!='$') {
			T= token_copy_deep(T);
			-[clist/insert T before (&value->tokens) ]
			continue;
		}
		arg_t *A= var_get(T->data, frame);
		if (!A) {
			T= token_copy_deep(T);
			-[clist/insert T before (&value->tokens) ]
			continue;
		}
		token_t *S;
		-[ for each S in list (&A->tokens) , begin ]
			S= token_copy_deep(S);
			-[clist/insert S before (&value->tokens) ]
		-[ end ]
	-[ end ]

	var_set(varname, value, frame);
	free(varname);
	return 0;
}


-[public] arg_t *arglist(arg_t *L,int nr) {
	int i=0;
	arg_t *a;
	-[ for each a in list L, begin ]
		if (i==nr)
			return a;
		i++;
	-[ end ]
	return 0;
}


arg_t *blt_def(arg_t *call, frame_t *_frame) {
	arg_t *body,*name_a;

	body= arglist(call,0);
	name_a= arglist(call,1);
	if (!body || !name_a)
		wrong_args("def");

	token_t *name= toklist(&name_a->tokens,0);
	if (!name || name->kind!='W' || name->data[0]=='$')
		wrong_args("def");

	char *err;
	stmt_t *stmts;
	err= parse_arg(body, &stmts);
	if (err)
		parse_error(name->data,err);

	-[ remove the old definition ]

	userfunc_t *U;
	-[ptr/new U]
	U->body= stmts;
	U->filename= file_name;

	function_t *F;
	-[ptr/new F]
	F->kind= 'U';
	F->func= U;
	F->name= strdup(name->data);
	strtab_insert(&functions, name->data, F);
	return 0;
}

=[ remove the old definition ] {
	function_t *F;
	F= strtab_find(&functions, name->data);
	if (F) {
		strtab_remove(&functions,name->data);
		if (F->kind=='U') {
			userfunc_t *U= F->func;
			stmtlist_destroy_deep(U->body);
			free(U);
		}
		free(F);
	}
}

=[]


-[private] typedef struct {
	stmt_t *body;
	char *filename;
} userfunc_t;

-[public] token_t *toklist(token_t *L,int nr) {
	int i=0;
	token_t *T;
	-[ for each T in list L, begin ]
		if (i==nr)
			return T;
		i++;
	-[ end ]
	return 0;
}

-[public] void wrong_args(char *fn) {
	exit(fprintf(stderr,"wrong number or kind of arguments to function `%s'\n",fn));
}

-[public] void parse_error(char *fn,char *msg) {
	exit(fprintf(stderr,"parse error in function %s: %s\n", fn,msg));
}

-[private] void
call_user_function(char *name,arg_t *call,userfunc_t *uf, frame_t *pframe) {
	stmt_t *actual;

	actual= stmtlist_copy_deep(uf->body);
	call= arglist_copy_deep(call);

	push_callstack(uf->filename, name);

	resolve_vars_in_arglist( call, pframe );

	resolve_parameters_in_body(actual,call);

	frame_t frame;
	memset(&frame,0,sizeof(frame));
	stmt_t *S;
	-[ for each S in list actual, begin ]
		arg_t *result;
		result= exec_stmt(S,&frame);
		if (result)
			arg_destroy_deep(result);
	-[ end ]
	-[ empty frame ]

	arglist_destroy_deep(call);
	stmtlist_destroy_deep(actual);

	pop_callstack();
}



-[private] void
resolve_vars_in_arglist(arg_t *call, frame_t *frame) {
	arg_t *A;
	-[ for each A in list call, begin ]
		token_t *T;
		-[ for each T in list (&A->tokens) , begin ]
			if (T->kind!='W')
				continue;
			if (T->data[0]!='$')
				continue;
			if (resolve_var(T, frame)) {
				-[ clist/remove T ]
			}
		-[ end ]
	-[ end ]
}

-[private] int resolve_var(token_t *T,frame_t *frame) {
	arg_t *value=0;
	if (islower(T->data[1]))
		value= frame_get(frame,T->data);
	if (!value)
		return 0;
	arg_t *copy= arg_copy_deep(value);
	token_t *C;
	-[ for each C in list (&copy->tokens) , begin ]
		-[clist/remove C]
		-[clist/insert C before T]
	-[ end ]
	free(copy);
	return 1;
}

-[private] void
resolve_parameters_in_body(stmt_t *body, arg_t *call) {
	stmt_t *S;
	-[ for each S in list body, begin ]
		arg_t *A;
		-[ for each A in list (&S->args) , begin ]
			token_t *T;
			-[for each T in list (&A->tokens) , begin ]
				if (T->kind!='W')
					continue;
				if (T->data[0]!='$')
					continue;
				if (!isdigit(T->data[1]))
					continue;
				resolve_parameter(T,call);
				-[ clist/remove T ]
			-[end]
		-[ end ]
	-[ end ]
}

-[private] void
resolve_parameter(token_t *T,arg_t *call) {
	int s;
	char *u;
	u= T->data+1;
	-[ calculate s ]
	arg_t *arg;
	arg= arglist(call, s);
	if (!arg)
		bad_parameter(T->data);
	if (!*u) {
		token_t *param;
		-[ for each param in list (&arg->tokens) , begin ]
			token_t *C;
			C= token_copy_deep(param);
			-[clist/insert C before T]
		-[ end ]
		return ;
	}
	if (*u!='.')
		exit(fprintf(stderr,"syntax error in positional parameter: %s\n",T->data));
	u++;
	-[ calculate s ]

	token_t *t;
	t= toklist(&arg->tokens, s);
	if (!t)
		bad_parameter(T->data);
	token_t *C;
	C= token_copy_deep(t);
	-[clist/insert C before T]
}

=[ calculate s ]
	s= 0;
	while(*u) {
		if (!isdigit(u[0]))
			break;
		s= s*10 + u[0]-'0';
		u++;
	}

=[]


-[private] void
bad_parameter(char *what) {
	exit(fprintf(stderr,"bad parameter %s: not so many parameters here\n",what));
}


=[ empty frame ] {
	arg_t *V;
	while((V=strtab_destroy(&frame.locals)))
		arg_destroy_deep(V);
}
=[]
	
-[public] arg_t* var_get(char *name,frame_t *frame) {
	if (!strcmp(name,"$HERE"))
		return here;
	if (!strcmp(name,"$CWD"))
		return working_dir;
	if (islower(name[1]))
		return frame_get(frame, name);
	else
		return frame_get(&global_frame, name);
}

-[public] frame_t global_frame;

-[public] void var_set(char *name,arg_t *value,frame_t *frame) {
	if (islower(name[1]))
		frame_set(frame, name,value);
	else
		frame_set(&global_frame, name,value);
}

arg_t* blt_comment(arg_t *call,frame_t *f) {
	/* FIXME: print the comment if we are in the trace mode */
	return 0;
}

arg_t *blt_listtrans(arg_t *call,frame_t *frame) {
	arg_t *body_a= arglist(call,0);
	arg_t *src_a = arglist(call,1);
	char *iter= get_varname(call, 2, "list.trans");

	arg_t *dst_a;
	char *dst;
	dst_a= arglist(call,3);
	if (!dst_a)
		dst= 0;
	else {
		token_t *T= toklist(&dst_a->tokens,0);
		if (T->kind!='W' || T->data[0]!='$')
			wrong_args("list.trans");
		dst= T->data;
	}

	if (!body_a || !src_a)
		wrong_args("list.trans");

	arg_t *nV;
	-[ptr/new nV]
	-[clist/init (&nV->tokens) ]

	token_t *src= expand_variables(&src_a->tokens, frame);
	token_t *W;
	-[ for each W in list src, begin ]
		token_t *T= token_copy_deep(W);
		if (W->kind!='W') {
			-[clist/insert T before (&nV->tokens) ]
		} else {
			arg_t *tempV= arg_new();
			-[ clist/insert T before (&tempV->tokens) ]
			/* frame_set does the deallocation so we dont 
			    need to remember tempV */
			frame_set(frame, iter, tempV);

			stmt_t *statements;
			char *err;
			err= parse_arg(body_a, &statements);
			if (err)
				parse_error("list.trans",err);
			stmt_t *S;
			arg_t *result=0;
			-[ for each S in list statements, begin ]
				arg_t *ret;
				ret= exec_stmt(S, frame);
				if (ret) {
					if (result)
						arg_destroy_deep(result);
					result= ret;
				}
			-[ end ]

			if (result) {
				-[ for each T in list (&result->tokens) , begin ]
					-[clist/remove T]
					-[clist/insert T before (&nV->tokens) ]
				-[ end ]
				free(result);
			}
		}
	-[ end ]

	if (dst)
		var_set(dst, nV, frame);
	else
		arg_destroy_deep(nV);
	return 0;
}

-[private] token_t *expand_variables(token_t *L, frame_t *frame) {
	token_t *T;
	token_t *nL= token_new();
	
	-[ for each T in list L, begin ]
		if (T->kind!='W' || T->data[0]!='$') {
			token_t *C= token_copy_deep(T);
			-[clist/insert C before nL]
			continue;
		}
		arg_t *A= var_get(T->data, frame);
		if (!A) {
			token_t *C= token_copy_deep(T);
			-[clist/insert C before nL]
		} else {
			token_t *S;
			-[ for each S in list (&A->tokens) , begin ]
				token_t *C= token_copy_deep(S);
				-[clist/insert C before nL]
			-[ end ]
		}
	-[ end ]

	return nL;
}

-[public] char *get_varname(arg_t *call,int idx,char *fname) {
	arg_t *A;
	token_t *T;
	A= arglist(call,idx);
	if (!A)
		wrong_args(fname);
	T= toklist(&A->tokens, 0);
	if (!T || T->kind!='W' || T->data[0]!='$')
		wrong_args(fname);
	return T->data;
}

-[private] typedef struct callstack {
	struct callstack *prev,*next;
	char *file;
	char *func;
} callstack_t;

-[private] callstack_t callstack;

-[public] void push_callstack(char *file,char *func) {
	callstack_t *C;
	-[ptr/new C]
	C->file= file;
	C->func= func;

	if (!callstack.prev)
		-[ clist/init (&callstack) ]
	-[clist/insert C before (&callstack) ]
}

-[public] void pop_callstack() {
	callstack_t *C;
	C= callstack.prev;
	if (C!=&callstack) {
		-[ clist/remove C]
		free(C);
	}
}

-[public] void print_callstack() {
	callstack_t *C;
	-[ for each C in list (&callstack) , begin ]
		fprintf(stderr,"file %s  function %s\n", C->file, C->func);
	-[ end ]
}

-[public] void runtime_error(char *fn,char *msg) {
	fprintf(stderr,"error in builtin function %s: %s\n",fn,msg);
	fprintf(stderr,"call stack:\n");
	print_callstack();
	exit(1);
}

-[ header "ctype.h" ]

-[public] void var_kill(char *name,frame_t *frame) {
	if (!strcmp(name,"$HERE"))
		return ;
	if (!strcmp(name,"$CWD"))
		return ;
	if (islower(name[1]))
		frame_kill(frame, name);
	else
		frame_kill(&global_frame, name);
}
