-[library "schedule.c", header "schedule.h" ]

#define MAXUNSIGNED 0xffffffffu

-[public] void
run_commands() {
	create_command_and_output_list();
	while(cmds.prev!=&cmds) {
		cmdlist_t base;
		calculate_base(&base);
		filter_clean_commands(&base);
		cmdlist_t *C;
		-[ for each C in list (&base) , begin ]
			execute_command(C);
			outputs_are_done(C);
			-[clist/remove C]
			cmd_destroy(C);
		-[ end ]
	}
}


-[private] void
create_command_and_output_list() {
	while(1) {
		strtab_t cmd;
		cmdlist_t *c;
		arg_t *O;
		if (command_get_one(&cmd))
			return ;
		-[ptr/new c]
		-[ private] typedef struct cmdlist {
			struct cmdlist *prev,*next;
			strtab_t fields;
		} cmdlist_t;
		+[private variables] static cmdlist_t cmds = {&cmds, &cmds}; +[ end]
		c->fields= cmd;
		-[clist/insert c before (&cmds) ]
		-[ private] strtab_t outputs;
		O= strtab_find(&cmd,"outputs");
		if (!O)
			continue;
		token_t *T;
		-[ for each T in list (&O->tokens) , begin ]
			if (T->kind!='W') continue;
			if (!strtab_find(&outputs,T->data))
				strtab_insert(&outputs, T->data, (void*) 1);
		-[ end ]
	}
}



-[ private ] void
calculate_base(cmdlist_t *head) {
	cmdlist_t *C;
	-[clist/init head]
	-[for each C in list (&cmds) , begin ]
		arg_t *triggers;
		triggers= strtab_find(&C->fields, "triggers");
		if (!triggers) {
			-[clist/remove C]
			cmd_destroy(C);
			continue;
		}
		token_t *T;
		int flag=0;
		-[ for each T in list (&triggers->tokens) , begin ]
			if (T->kind!='W')
				continue;
			if (strtab_find(&outputs,T->data)) {
				flag=1;
				break;
			}
		-[ end ]
		if (!flag) {
			-[clist/remove C]
			-[clist/insert C before head]
		}
	-[end ]
}

-[private] void filter_clean_commands(cmdlist_t *L) {
	cmdlist_t *C;
	-[ for each C in list L, begin ]
		if (command_is_clean(C)) {
			-[clist/remove C]
			are_outputs_dirty(C);
			outputs_are_done(C);
			cmd_destroy(C);
		}
	-[ end ]
}


-[private] int command_is_clean(cmdlist_t *C) {
	unsigned min_o= MAXUNSIGNED,max_i=0;
	arg_t *A;
	token_t *T;

	A= strtab_find(&C->fields,"triggers");
	if (!A) 
		return 0;
	-[for each T in list (&A->tokens) , begin ]
		if (T->kind!='W') continue;
		unsigned mt= modtime_input(T->data);
		if (mt>max_i)
			max_i= mt;
	-[ end ]

	A= strtab_find(&C->fields, "outputs");
	if (!A)
		return 0;
	-[ for each T in list (&A->tokens) , begin ]
		if (T->kind!='W') continue;
		unsigned mt= modtime_output(T->data);
		if (mt<min_o)
			min_o= mt;
	-[ end ]
	return max_i <= min_o;
}

-[private] void cmd_destroy(cmdlist_t *C) {
/*	arg_t *A;
	while((A=strtab_destroy(&C->fields)))
		arg_destroy_deep(A);*/
	free(C);
}

-[private] void
execute_command(cmdlist_t *C) {
	char *cmd;
	char *msg=0;
	cmd= prepare_string(C,"action");
	if (!cmd) 
		return ;
	if (option_printmsg) 
		msg= prepare_string(C,"message");
	if (!option_noprefix)
		printf("jj: ");
	printf("%s\n", msg ? msg : cmd );
	if (option_norun) {
		outputs_are_very_dirty(C);
		return ;
	}
	if (system(cmd))
		exit(fprintf(stderr,"jj: child failed. exiting.\n"));
	are_outputs_dirty(C);
}

-[private] char*
prepare_string(cmdlist_t *C,char *field) {
	arg_t *A;
	A= strtab_find(&C->fields, field);
	if (!A)
		return 0;

	char *result;
	int len;
	len= prep_string(A, C, 0);
	if (len<2)
		return 0;
	result= malloc(len);
	prep_string(A,C,result);
	return result;
}


-[private] int 
prep_string(arg_t *arg, cmdlist_t *C, char *result) {
	token_t *T;
	int len=0;
	-[ for each T in list (&arg->tokens), begin ]
		if (T->kind!='W')  {
			len= put_token(result,len,T->kind);
		} else if (T->data[0]!='$') {
			len= put_string( result, len, T->data );
		} else {
			arg_t *A;
			A= strtab_find(&C->fields, T->data+1);
			if (A) {
				token_t *S;
				-[ for each S in list (&A->tokens) , begin ]
					if (S->kind=='W')
						len= put_string(result,len,S->data);
					else
						len= put_token(result,len,S->kind);
				-[ end ]
			} else {
				len= put_string( result, len, T->data );
			}
		}
	-[ end ]
	if (result)
		result[len]= 0;
	return len+1;
}

-[ private] int put_string(char *result, int len, char *src) {
	int L= strlen(src);
	if (result) {
		memcpy(result+len, src, L);
		result[len+L]= ' ';
	}
	len+= L+1;
	return len;
}

-[private] int put_token(char *result, int len, int tk) {
	char buf[2];
	buf[0]= tk;
	buf[1]= 0;
	return put_string(result, len, buf);
}

-[private] void
outputs_are_very_dirty(cmdlist_t *C) {
	arg_t *A;
	A= strtab_find(&C->fields, "outputs");
	if (!A)
		return ;
	token_t *T;
	-[ for each T in list (&A->tokens) , begin ]
		if (T->kind=='W')
			modtime_force(T->data,MAXUNSIGNED);
	-[ end ]
}

-[private] void
are_outputs_dirty(cmdlist_t *C) {
	arg_t *A;
	A= strtab_find(&C->fields, "outputs");
	if (!A)
		return ;
	token_t *T;
	-[ for each T in list (&A->tokens) , begin ]
		if (T->kind=='W')
			modtime_update(T->data);
	-[ end ]
}

-[private] void
outputs_are_done(cmdlist_t *C) {
	arg_t *A;
	A= strtab_find(&C->fields, "outputs");
	if (!A)
		return ;
	token_t *T;
	-[ for each T in list (&A->tokens) , begin ]
		if (T->kind=='W')
			strtab_remove(&outputs,T->data);
	-[ end ]
}


-[private] unsigned
modtime_input(char *fn) {
	unsigned mt;
	mt= modtime(fn);
	if (!mt)
		exit(fprintf(stderr,"can not stat %s: %s\n",fn,strerror(errno)));
	return mt;
}

-[private] unsigned
modtime_output(char *fn) {
	return modtime(fn);
}

-[private] unsigned modtime(char *fn) {
	unsigned V;
	V= (unsigned) strtab_find(&modtimes,fn);
	if (V)
		return V;
	struct stat sbuf;
	if (stat(fn,&sbuf))
		return 0;
	V= sbuf.st_mtime;
	if (!V)
		V++;
	strtab_insert(&modtimes, fn, (void*) V);
	return V;
}

-[private] void modtime_force(char *fn,unsigned V) {
	if (strtab_find(&modtimes,fn))
		strtab_remove(&modtimes,fn);
	strtab_insert(&modtimes, fn, (void*) V);
}

-[private] strtab_t modtimes;

-[private] void modtime_update(char *fn) {
	unsigned V;
	V= (unsigned) strtab_find(&modtimes,fn);
	if (V)
		strtab_remove(&modtimes,fn);
	struct stat sbuf;
	if (stat(fn,&sbuf)) {
		strtab_insert(&modtimes,fn, (void*) MAXUNSIGNED);
		return ;
	}
	V= sbuf.st_mtime;
	if (!V)
		V++;
	strtab_insert(&modtimes, fn, (void*) V);
}

-[ lheader "parser.h" ] -[ lheader "table.h" ]
-[ lheader "main.h" ]
-[ header "unistd.h" ] -[header "stdlib.h"] -[ header "stdio.h" ]
-[ header "errno.h" ] -[ header "string.h"]
-[ header "sys/stat.h" ] -[lheader "command.h" ]


