-[  c library "gccdeps.c", header "gccdeps.h" ]

/* usage: {dirname} depdir */
arg_t *blt_depdir(arg_t *call,frame_t *frame) {
	arg_t *darg;
	char *dname;
	darg= arglist(call,0);
	if (!darg)
		wrong_args("depdir");
	dname= arg_to_string(darg,frame);
	read_deps_from_dir(dname);
	free(dname);
	return 0;
}

-[headers, begin ]
	!stdio !stdlib
-[end]
-[ lheader "table.h" ]
-[ lheader "parser.h" ]
-[ lheader "functions.h" ]
-[ lheader "string.h" ]
-[ header "dirent.h" ]
-[ lheader "fileio.h" ] -[ lheader "extlibs.h" ] -[header "stdarg.h"]


-[private] void read_deps(char *fname) {
	char *buf;
	char *target;
	char *part;
	char *sources,*targets;

	int pno;
	buf= file_read_all(fname);
	if (!buf)
		return ;
	shrink_pairs(buf,"\\\n",' ');
	replace_range(buf, 1, ' ', ' ');
	while(shrink_pairs(buf,"  ", ' '))
		;
	pno= 0;
	-[ for each part in string buf / ':' , begin ]
		if (pno)
			sources= part;
		else
			targets= part;
		pno++;
		part= 0;
	-[ end ]
	free(buf);
/*	printf("sources= <%s>\n",sources);
	printf("targets= <%s>\n",targets);*/
	if (!sources || !targets)
		return ;

	-[ for each target in string targets / ' ', begin ]
		char *v;
		if (!target)
			-[ continue ]
		v= strtab_find(&gcc_deps, target);
		if (v) {
			char *u= mprintf("%s %s",v,sources);
			strtab_remove(&gcc_deps, target);
			free(v);
			v= u;
		} else v= strdup(sources);
		strtab_insert(&gcc_deps, target, v);
	-[ end ]
	free(sources);
	free(targets);
}

-[public] char *get_gcc_deps(char *target) {
	char *deps;
	deps=  strtab_find(&gcc_deps, target);
	return deps;
}

-[private] strtab_t gcc_deps;


-[private] int shrink_pairs(char *buf,char *p,int rep) {
	char *dst;
	int change= 0;
	dst= buf;

initial:
	if (buf[0]==0) {
		*dst= 0;
		return change;
	}

	if (buf[0]!=p[0]) {
		buf++;
		dst++;
		goto initial;
	}

	if (buf[1]!=p[1]) {
		buf++;
		dst++;
		goto initial;
	}

	dst[0]= rep;
	buf+= 2;
	dst++;
	change= 1;
	goto initial;
}

-[private] void replace_range(char *buf,int lower, int upper, int rep) {
	int i;
	for(i=0;buf[i];i++)
		if (buf[i]>=lower && buf[i]<=upper)
			buf[i]= rep;
}

-[private] void read_deps_from_dir(char *dname) {
	DIR *dir= opendir(dname);
	struct dirent *dirent;
	if (!dir)
		return ;
	while((dirent=readdir(dir))) {
		char buf[2000];
		if (has_suffix(dirent->d_name, ".d")) {
			sprintf(buf,"%s/%s",dname, dirent->d_name);
		 	read_deps(buf);
		}
	}
	closedir(dir);
}

-[private] int has_suffix(char *str,char *sfx) {
	int La= strlen(str);
	int Lb= strlen(sfx);
	if (La<Lb)
		return 0;
	return !strcmp(str+La-Lb, sfx);
}