-[ library "main.c" , header "main.h" ]

-[ public] char *file_name;

-[private] void
load_file(char* fname, arg_t *rel) {
	char *xoldf;
	stmt_t *statements;
	char *err;
	stmt_t *S;

	if (fname[0]!='/') 
		fname= relative_path(fname, rel->tokens.next->data);
	fname= normalize(fname);
	xoldf= file_name;
	file_name= fname;
	compute_here();
	err= parse_file(fname,&statements);
	if (err)
		exit(fprintf(stderr,"parse error in file %s: %s\n",fname,err));
	
	push_callstack(fname, "(top)");
	-[ for each S in list statements, begin ]
		arg_t *result;
		result= exec_stmt(S, &global_frame);
		if (result)
			arg_destroy_deep(result);
	-[ end ]
	pop_callstack();
	file_name= xoldf;
	if (file_name)
		compute_here();
}

-[private] char *relative_path(char *fname,char *rel) {
	char *R;
	int fL= strlen(fname);
	int rL= strlen(rel);
	R= malloc(rL+fL+2);
	memcpy(R, rel, rL);
	R[rL]= '/';
	memcpy(R+rL+1, fname, fL);
	R[rL+fL+1]= 0;
	return R;
}

-[private] char *
normalize(char *fn) {
        char *x;
        while((x= strstr(fn,"//")))
                memmove(x,x+1,(strlen(x+1)+1)*sizeof(char));

        while((x=strstr(fn,"/./")))
                memmove(x,x+2,(strlen(x+2)+1)*sizeof(char));

        while((x=strstr(fn,"/../"))) {
                if (x==fn)  {
                        memmove(fn,fn+3,(strlen(x+3)+1 )*sizeof(char));
                } else {
                        char *up;
                        for(up=x-1;*up!='/';up--)
                                ;
                        memmove(up,x+3,(strlen(x+3)+1)*sizeof(char));
                }
        }
        return fn;
}


-[public] arg_t *here;
-[public] arg_t *working_dir;



void main_init() {
	static char wdstore[2000];
	char *wdir= getcwd(wdstore,1990);
	if (!wdir)
		exit(fprintf(stderr,"such long working directories confuse me\n"));
	strcat(wdir,"/");
	if (working_dir)
		arg_destroy_deep(working_dir);
	working_dir= arg_new();
	token_t *T= token_new();
	T->kind= 'W';
	T->data= strdup(wdir);
	-[clist/insert T before (&working_dir->tokens)]
	init_builtin_functions();
}

-[private] void
compute_here() {
	int i;
	char *fdir= strdup(file_name);
	int fL= strlen(fdir);
	for(i=fL-1;i&& fdir[i]!='/';i--)
		;
	fdir[i+1]= 0;
	if (here)
		arg_destroy_deep(here);
	here= arg_new();
	token_t *T= token_new();
	T->kind= 'W';
	T->data= fdir;
	-[clist/insert T before (&here->tokens) ]
}

-[public] int option_printmsg;

int main(int argc,char **argv) {
	int i;
	int loaded= 0;
	main_init();
	for(i=1;i<argc;i++) {
		-[ flag argument "-m" -> option_printmsg ]
		-[ flag argument "-n" -> option_norun ]
		-[ flag argument "-p" -> option_noprefix ]
		-[ flag argument "-g" -> option_graph ]
		-[ flag argument "-q" -> option_quiet ]
		if (argv[i][0]!='-') {
			load_file(argv[i], working_dir);
			loaded++;
		}
	}
	if (!loaded)
		load_file("jjcool", working_dir);
	if (option_graph)
		dump_dependency_graph();
	else
		run_commands();
	return 0;
}

=[ flag argument $STR -> $V ]
+[ public variables ] extern int $V; +[end]
+[ private variables] int $V= 0; +[end]
	if (!strcmp(argv[i],$STR)) {
		$V= 1;
		continue;
	}
=[]

arg_t *blt_include(arg_t *call, frame_t *frame) {
	arg_t *A;
	-[ for each A in list call, begin ]
		include_file(A, frame);
	-[ end ]
	return 0;
}

-[private] void include_file(arg_t* arg,frame_t *frame) {
	char *name;
	int len;
	len= stringcat(arg, 0, frame);
	name= malloc(len);
	stringcat(arg,name, frame);
	load_file(name,here);
}

-[ lheader "parser.h" ]
-[ lheader "ddepg.h" ]
-[ header "stdlib.h" ]
-[ header "stdio.h" ]
-[ header "string.h" ]
-[ lheader "table.h" ]
-[ lheader "functions.h" ]
-[ lheader "string.h" ]
-[ header "unistd.h" ]
-[ lheader "schedule.h" ]


