
-[ sstack_t : private stack of char* ]

#define LCURLY '{'
#define RCURLY '}'

-[public] int ipret_block(char *fn, char *input)
{
	int state;
	sstack_t *stack;
	int Z;
	char *t_begin;
	int ttype;
	char *val;
	char *str;

	stack= sstack_new();
	state= 0;

	while(1) {
	-[ tuple switch state X *input, begin ]
	-[case 0 X 0 ] goto end;
	-[caser 0 X 1 .. 32 ]
	-[case 0 X 127 ] input++; break;
	-[case 0 X LCURLY ] Z= 1; -[ token transition 'C']
	-[case 0 X RCURLY] 
		do_error(fn, input,  "unmatched curly brace");
		break;
	-[case 0 X '$' ] ttype='V'; -[ token transition 'W' ]
	-[case 0 X '#' ] ttype='F'; -[ token transition 'W']
	-[cased 0 ] ttype='S'; -[ token transition 'W']

	-[case 'C' X LCURLY ]  Z++; input++; break;
	-[case 'C' X RCURLY ]
		input++;
		Z--;
		if (Z==0) {
			t_begin++;
			sstack_push(stack,
				make_str(t_begin, input-t_begin-1));
			state=0;
		}
		break;
	-[case 'C' X 0 ]
		do_error(fn, 0,"unexpected EOF in string");
		break;
	-[cased 'C' ] input++; break;

	-[caser 'W' X 1 .. 32]
	-[case 'W' X RCURLY]
	-[case 'W' X LCURLY]
		str= make_str(t_begin,input-t_begin);
		switch(ttype) {
		case 'S':
			sstack_push(stack,str);
			str= 0;
			break;
		case 'F':
			call_function(str,stack);
			break;
		case 'V':
			val= var_get(str);
			if (val) {
				sstack_push(stack, strdup(val));
			} else {
				sstack_push(stack,str);
				str= 0;
			}
			break;
		}
		if (str) free(str);
		state= 0;
		break;
	-[cased 'W'] input++; break;

	-[ dcase 0 ]
		do_error(fn, 0,"unexpected EOF in statement");
		break;
	-[ end ]
	}
end:
	empty_stack(stack);
	sstack_destroy(stack);
	free(stack);
	return 0;
}



-[ import file name functions]
-[ import file_read_all in private scope ]
-[ import mprintf library ]


=[ token transition $STATE ]
	t_begin= input;
	input++;
	state= $STATE;
	break;

=[]

-[private] void do_error(char *fn,char *ref,char *msg) {
	fprintf(stderr,"%s: %s\n",fn,msg);
	exit(1);
}

-[private] char *make_str(char *b,int L)
{
	char *R;
	R= malloc(L+1);
	memcpy(R,b,L);
	R[L]= 0;
	return R;
}

-[public] int ipret_file
	(char *dir,char *fn)
{

	char *input;
	char *here_old;
	int ret;

	if (fn[0]!='/') 
		fn= fs_normalize(mprintf("%s/%s",dir,fn));
	else
		fn= fs_normalize(strdup(fn));
	input= file_read_all(fn);
	if (!input)
		exit(fprintf(stderr,"can not read file %s: %s\n",
			fn,strerror(errno)));
	here_old= strdup(var_get("$HERE"));
	var_set("$HERE", fs_dirname(fn));
	ret=  ipret_block(fn,input);
	var_set("$HERE", here_old);
	return ret;
}

-[private] void call_function
	(char *name,sstack_t *stack)
{
	func_t *F;

	F= func_get(name);
	if (!F)
		exit(fprintf(stderr,"can not find function %s\n", name));

	if (F->kind==fn_user) 
		expand_args(stack);

	frame_enter();
	bind_args( &F->args, stack );

	if (F->kind==fn_user) {
		ipret_block("<internal>", F->data);
	} else {
		void (*func)();
		func= F->data;
		func();
	}
	
	frame_leave();
	empty_stack(stack);
}


-[private] void bind_args
	(strlist_t *arglist, sstack_t *stack)
{
	int i;
	i= 0;
	-[ list>for each argname in *arglist , begin ]
		char *val;
		if (i>=stack->nelts)
			val= "";
		else
			val= stack->elts[i];
		var_create(argname, val);
		i++;
	-[ end ]
}

-[private] void expand_args(sstack_t *stack)
{
	int i;
	for(i=0;i<stack->nelts;i++) {
		char *X,*Y;
		X= stack->elts[i];
		Y= expand_variables_in_str(X);
		if (Y!=X) {
			free(stack->elts[i]);
			stack->elts[i]= Y;
		}
	}
}

-[ charbuffer_t : private buffer [500] of char ]

-[private] char* expand_variables_in_str
	(char *input)
{
	charbuffer_t *buffer;
	int size;

	if (!strchr(input,'$'))
		return input;

	buffer= charbuffer_new();

	while(*input) {
		char *begin;
		char *s;
		char *val;
		if (*input!='$') {
			charbuffer_append(buffer, input[0]);
			input++;
			continue;
		}
		begin= input;
		while(*input
			&& *input!=127 && (*input<0 || *input>32)
			&& *input!=LCURLY && *input!=RCURLY) {
				input++;
		}
		s= make_str(begin,input-begin);
		val= var_get(s);
		free(s);
		if (val) {
			for(s=val;*s;s++)
				charbuffer_append(buffer,s[0]);
		} else {
			while(begin!=input) {
				charbuffer_append(buffer, begin[0]);
				begin++;
			}
		}
	}
	charbuffer_append(buffer,0);
	
	return charbuffer_collectda(buffer, &size);
}


-[ unused symbol sstack_init ]
-[ unused symbol charbuffer_collectd ]

-[private] void empty_stack(sstack_t *stack) {
	while(!sstack_isempty(stack)) {
		char *S;
		S= sstack_pop(stack);
		free(S);
	}
	stack->nelts= 0;
}
